import numpy as np
import editdistance
from scipy.stats import wasserstein_distance

#TODO add more

def makeLn(n):
    return lambda a, b: np.linalg.norm(a - b, ord=n)

def editDistance(a, b):
    return editdistance.eval(a, b)

def earthMoverDistance(u, v, u_weights=None, v_weights=None):
    """
    Compute the first Wasserstein distance between two 1D distributions.
    :param u, v: (Array-like) Values observed in the (empirical) distribution.
    :param u_weights, v_weights: (Array-like, optional) Weight for each value. If left unspecified,
    each value is assigned the same weight. u_weights (resp. v_weights) must have the same length
    as u_values (resp. v_values). If the weight sum differs from 1,
    it must still be positive and finite so that the weights can be normalized to sum to 1.
    :return: (float) The computed distance between the distributions.
    """
    return wasserstein_distance(u, v, u_weights, v_weights)



import os
import sys
import numpy as np

from time                     import time
from sklearn.neighbors        import KNeighborsClassifier
from sklearn.metrics.pairwise import pairwise_distances

sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'nn_condensing', 'Python Implementation'))
from nn_condensing import nn # this only looks like an error because the IDE doesn't understand the ugly hack above ^
from Utils         import computeGram, \
                          computeGammaSet, \
                          computeLabels, \
                          computeAlpha, \
                          computeQ


def constructGammaNet(Xs, gram, gamma, prune):
    chosenXs = nn.epsilon_net_hierarchy(data_sample=Xs,
                                        epsilon=gamma,
                                        distance_measure=None,
                                        gram_matrix=gram)

    if prune:
        chosenXs = nn.consistent_pruning(net=chosenXs,
                                         distance_measure=None,
                                         gram_matrix=gram)

    return chosenXs

class KSU(object):

    def __init__(self, Xs, Ys, gramPath, metric, logger, prune=False):
        self.classifier = None
        self.Xs         = Xs
        self.Ys         = Ys
        self.prune      = prune
        self.logger     = logger
        self.metric     = metric

        if gramPath is None:
            self.logger.info('Computing Gram matrix...')
            tStartGram = time()
            self.gram  = pairwise_distances(self.Xs, metric=self.metric, n_jobs=-1)
            self.logger.debug('Gram computation took {:.3f}s'.format(time() - tStartGram))
        else:
            self.logger.info('Loading Gram matrix from file...')
            self.gram = np.load(gramPath) # TODO if we change from numpy, change here

    def predict(self, x):
        if self.classifier is None:
            raise RuntimeError("Predictor not generated yet. you must run KSU.makePredictor() before predicting")
        else:
            return self.classifier.predict(x)

    def makePredictor(self, delta):
        gammaSet = computeGammaSet(self.gram)
        qMin     = float(np.inf)
        n        = len(self.Xs)

        self.logger.debug('Choosing from {} gammas'.format(len(gammaSet)))
        for gamma in gammaSet:
            tStartGamma = time()
            gammaXs     = constructGammaNet(self.Xs, self.gram, gamma, self.prune)
            tStartLabel = time()
            gammaYs     = computeLabels(gammaXs, self.Xs, self.Ys, self.gram, self.metric)
            alpha       = computeAlpha(gammaXs, gammaYs, self.Xs, self.Ys)
            m           = len(gammaXs)
            q           = computeQ(n, alpha, 2 * m, delta)

            self.logger.debug(
                'For gamma: {g}, net construction took {nt:.3f}s, label choosing took {lt:.3f}s, q: {q}'.format(
                    g=gamma,
                    q=q,
                    nt=tStartLabel - tStartGamma,
                    lt=time() - tStartLabel))

            if q < qMin:
                qMin      = q
                bestGamma = gamma
                chosenXs  = gammaXs
                chosenYs  = gammaYs

        self.logger.info('Chosen best gamma: {g}, which achieved q: {q}'.format(
            g=bestGamma,
            q=qMin))

        self.classifier = KNeighborsClassifier(n_neighbors=1, metric=self.metric, algorithm='auto', n_jobs=-1)
        self.classifier.fit(chosenXs, chosenYs)










import unittest
import sys


class TestComputeGram(unittest.TestCase):

    @staticmethod
    def suite():
        test_suite = unittest.TestSuite()
        test_suite.addTest(unittest.makeSuite(TestComputeGram))
        return test_suite

    def empty_elements(self):
        self.fail()

    def empty_metric(self):
        self.fail()

    def first(self):
        self.fail()

if __name__ == "__main__":
    # So you can run tests from this module individually.
    sys.exit(unittest.main())



import numpy as np
import time

from sklearn.neighbors import KNeighborsClassifier
from collections       import Counter
from math              import sqrt, log

def getDateTime():
    return time.strftime('%d.%m.%y %H:%M:%S')

def parseInputData(dataPath):
    data = np.load(dataPath)
    return data # TODO return as dict (like in w2v-parseNPZ)

def log2(x):
    return log(x, 2)

def computeGram(elements, dist): #unused
    n    = len(elements)
    gram = np.array((n, n))
    for i in range(n):
        for j in range(n - i):
            gram[i,j] = dist(elements[i], elements[j])

    lowTriIdxs       = np.tril_indices(n) #TODO make sure gram is upper triangular after the loop
    gram[lowTriIdxs] = gram.T[lowTriIdxs]

    return gram

def computeQ(n, m, alpha, delta):
    firstTerm  = (n * alpha) / (n - m)
    secondTerm = (m * log2(n) - log2(delta)) / (n - m)
    thirdTerm  = sqrt(((n * m * alpha * log2(n)) / (n - m) - log2(delta)) / (n - m))

    return firstTerm + secondTerm + thirdTerm

def computeLabels(gammaXs, Xs, Ys, gram, metric): # TODO deprecate after testing optimizedComputeLabels
    gammaYs = range(len(gammaXs))
    h = KNeighborsClassifier(n_neighbors=1, metric=metric, algorithm='auto', n_jobs=-1)
    h.fit(gammaXs, gammaYs)

    groups = {i:Counter() for i in gammaYs}
    for x, y in zip(Xs, Ys):
        groups[h.predict(x)].update(y)

    return [c.most_common(1)[0][0] for c in groups.keys()]

def computeAlpha(gammaXs, gammaYs, Xs, Ys, metric):
    classifier = KNeighborsClassifier(n_neighbors=1, metric=metric, algorithm='auto', n_jobs=-1)
    classifier.fit(gammaXs, gammaYs)
    return classifier.score(Xs, Ys)

def computeGammaSet(gram):
    gammaSet = np.unique(gram)
    gammaSet = np.delete(gammaSet, 0)
    return gammaSet

def findIndices(array, elements):
    return filter(lambda x: x is not None, [array.index(e) if e in array else None for e in elements])

def optimizedComputeLabels(gammaXs, Xs, Ys, gram):
    m                = len(gammaXs)
    n                = len(Xs)
    slice            = findIndices(Xs, gammaXs)
    gammmaGram       = gram[:, slice]
    flatGram         = np.reshape(gammmaGram, [-1])
    perm             = np.argsort(flatGram)
    flatYs           = np.array(Ys * m)[perm]
    flatXIdxs        = np.array([j for l in [[e] * m for e in range(m)] for j in l])[perm]
    flatNeighborIdxs = np.array(range(n) * m)[perm]
    taken            = np.full([n], False, dtype=bool)

    numTaken = 0
    groups   = {i: Counter() for i in range(m)}
    for y, xIdx, neighborIdx in zip(flatYs, flatXIdxs, flatNeighborIdxs):
        if numTaken == n:
            break
            
        if taken[neighborIdx]:
           continue

        groups[xIdx].update(y)
        taken[neighborIdx] = True
        numTaken += 1

    return [c.most_common(1)[0][0] for c in groups.keys()]



from unittest import TextTestRunner

from TestMetrics import TestEarthMoverDistance

if __name__ == "__main__":
    """
    Add unit test class names to testCases
    """
    testCases = [TestEarthMoverDistance]
    runner=TextTestRunner()

    for testCase in testCases:
        suite = testCase.suite()
        runner.run(suite)



# -*- coding: utf-8 -*-
""" install script
    thanks to Kenneth Reitz https://github.com/kennethreitz
    for publishing this setup.py template
"""
# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools.command.install import install
from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'ksu'
DESCRIPTION = 'Implementation of the KSU compression algorithm https://www.cs.bgu.ac.il/~karyeh/compression-arxiv.pdf'
URL = 'https://github.com/nimroha/ksu_classifier'
EMAIL = 'nimrod.morag@gmail.com'
AUTHOR = 'Nimrod Morag, Yuval Nissan'
REQUIRES_PYTHON = '>=2.7.0'
VERSION = None
REQUIRED = None

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '__version__.py')) as f:
        exec (f.read(), about)
else:
    about['__version__'] = VERSION

if not REQUIRED:
    with io.open(os.path.join(here, 'requirements.txt')) as f:
        REQUIRED = f.read().split('\n')

class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()

def cloneSubmodules(): #TODO only works in project root dir (TODO doesn't seem to clone)
    os.system('git submodule init')
    os.system('git submodule update --recursive')  # get the epsilon-net creation module from github

class CustomInstallCommand(install):
    """Add custom functionality to install command."""

    @staticmethod
    def _called_from_setup(run_frame): # hack to force backward-compatible install
        return True

    def run(self):
        self.execute(cloneSubmodules, (), msg='Fetching 1-nn-condensing submodule')
        install.run(self) # run regular install

# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['ksu'],

    entry_points={
        'console_scripts': ['ksu=ksu.EntryPoint:main'],
    },
    # scripts=['ksu/EntryPoint.py'],

    install_requires=REQUIRED,
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Development Status :: 4 - Beta'
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
        # 'install': CustomInstallCommand,
    },
)







import unittest
import sys

from ksu.Metrics import earthMoverDistance

class TestEarthMoverDistance(unittest.TestCase):

    @staticmethod
    def suite():
        test_suite = unittest.TestSuite()
        test_suite.addTest(unittest.makeSuite(TestEarthMoverDistance))
        return test_suite

    def test_no_weights(self):
        u = [0, 1, 3]
        v = [5, 6, 8]
        d = earthMoverDistance(u, v)
        self.assertEqual(d, 5.0)

    def test_with_weights_same_domain(self):
        u = [0, 1]
        v = [0, 1]
        u_weights = [3, 1]
        v_weights = [2, 2]
        d = earthMoverDistance(u, v, u_weights, v_weights)
        self.assertEqual(d, 0.25)

    def test_with_weights_different_domains(self):
        u = [3.4, 3.9, 7.5, 7.8]
        v = [4.5, 1.4]
        u_weights = [1.4, 0.9, 3.1, 7.2]
        v_weights = [3.2, 3.5]
        d = earthMoverDistance(u, v, u_weights, v_weights)
        self.assertEqual(d, 4.078133143804786)

if __name__ == "__main__":
    # So you can run tests from this module individually.
    sys.exit(unittest.main())



import sys
import argparse
import logging

from sklearn.neighbors.base import VALID_METRICS

import Metrics
from Utils import parseInputData
from KSU   import KSU

METRICS = {v:v for v in VALID_METRICS['brute'] if v != 'precomputed'}
METRICS['EditDistance'] = Metrics.editDistance
METRICS['EarthMover']   = Metrics.earthMoverDistance

def main(argv=None):

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description='Generate a KSU classifier')
    parser.add_argument('--data',          help='Path to input data file (in space separated key value format)',               required=True)
    parser.add_argument('--metric',        help='Metric to use (unless custom_metric is provided). {}'.format(METRICS.keys()), default='l2')
    parser.add_argument('--custom_metric', help='Absolute path to a directory (containing __init__.py) with a python file'
                                                'named Distance.py with a function named "dist(a, b)" that computes'
                                                'the distance between a and b by any metric of choice',                        default=None)
    parser.add_argument('--gram',          help='Path to a precomputed gram matrix (in .npz format)',                          default=None) # TODO decide which format. npz/panda/csv?
    parser.add_argument('--delta',         help='Delta parameter',                                                             default=0.1, type=float)  # TODO explain better, choose a good default
    parser.add_argument('--log_level',     help='Logging level',                                                               default='INFO')

    args = parser.parse_args()

    logging.basicConfig(level=args.log_level, filename='ksu.log')
    logger = logging.getLogger('KSU')
    logger.addHandler(logging.StreamHandler(sys.stdout))

    dataPath     = args.data
    gramPath     = args.gram
    metric       = args.metric
    delta        = args.delta
    customMetric = args.custom_metric

    if customMetric is not None:
        sys.path.append(customMetric)
        try:
            from Distance import dist  # this only looks like an error because we're importing from unknown path
            metric = dist
            logger.debug('Loaded dist function successfully')
        except:
            raise RuntimeError(
                'Could not import dist function from {p}'
                'make sure Distance.py and __init__.py exist in {p}'
                'and that Distance.py has a function dist(a, b)'.format(p=customMetric))
    else:
        if metric not in METRICS.keys():
            raise RuntimeError(
                '"{m}" is not a built-in metric. use one of {ms}'
                'or provide a custom metric with the --custom_metric argument'.format(
                    m=metric,
                    ms=METRICS.keys()
                ))

    logger.info('Reading data...')
    data = parseInputData(dataPath)

    ksu = KSU(data['X'], data['Y'], gramPath, metric, logger)
    h   = ksu.makePredictor(delta)




if __name__ == '__main__' :
    sys.exit(main())
